
%{

//standard includes
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdbool.h>
#include <unistd.h>

//my own headers
#include "regex.h"
#include "automata.h"
#include "xmlfsaparser.h"
#include "stringtester.h"
#include "minFSA.h"
#include "complement.h"
#include "rev.h"

//variables for parsing from the command-line
char *theInput;
char *theInputPtr;
char *theInputLim;

//what happens if there's an error
void yyerror(const char *str) {
	fprintf(stderr,"error: %s\n", str);
	exit(1);
}
 
//ignore line breaks; not used
int yywrap() { return 1; } 

//the regular expression produced by yyparse()
struct regexp *theRE;

%}

//tokenizer returns strings
%union {
	struct regexp *s;
	char c;
}

//strings are passed up the hierarchy
%token <c> SYMBOL
%token STAR LPAR RPAR TIE
%type <s> re une sre concat bre kleene ere group

%%

whole:	re {
				theRE = $1;
			}
			;

re:		une {
				$$ = $1;
			}
			| sre {
				$$ = $1;
			};

une:		re TIE sre {
				$$ = makeUnion($1,$3);
			}
			;

sre:		concat {
				$$ = $1;
			}
			| bre {
				$$ = $1;
			}
			;

concat:	sre bre {
				$$ = makeConcat($1,$2);
			}
			;

bre:		kleene {
				$$ = $1;
			}
			| ere {
				$$ = $1;
			}
			;

kleene:	ere STAR {
				$$ = makeKleene($1);
			}
			;

ere:		group {
				$$ = $1;
			}
			|
			SYMBOL {
				$$ = makeSimple($1);
			}
			;

group:	LPAR re RPAR {
				$$ = $2;
			}
			;

%%

int main(int argc, char *argv[]) {
	//initialize all the variables and flags
	theRE = NULL;
	char c;
	bool min = false;
	bool str = false;
	bool fsa = false;
	bool reg = false;
	bool eps = false;
	bool pfsa = false;
	bool det = false;
	bool xml = false;
	bool in = false;
	bool rev = false;
	bool comp = false;
	bool brzo = false;
	bool weights = false;
	char *filename = NULL;
	char *testString = NULL;
	struct fsa *f = NULL;
	//set the flags based on what options are given
	while ((c = getopt(argc, argv, "MRwcmxdpefr:i:s:")) != -1) {
		//check if program has to make an FSA
		if (c == 'f') {
			fsa = true;
		//check if weights should be printed
		} else if (c== 'w') {
			weights = true;
		//check if FSA is to be minimized by Brzozowski algorithm
		} else if (c == 'M') {
			brzo = true;
		//check if FSA is to be reversed
		} else if (c == 'R') {
			rev = true;
		//check if complement is to be constructed
		} else if (c == 'c') {
			comp = true;
		//check if the user has entered a regular expression
		} else if (c == 'r') {
			reg = true;
			theInput = strdup(optarg);
		//check if a particular string is well-formed
		} else if (c == 's') {
			str = true;
			testString = strdup(optarg);
			fprintf(stderr, "test string = \"%s\"\n", testString);
		//check if the user has entered an XML file name
		} else if (c == 'i') {
			in = true;
			filename = strdup(optarg);
			fprintf(stderr, "filename = %s\n", filename);
		//check if epsilon arcs should be removed
		} else if (c == 'e') {
			eps = true;
		//check if the FSA is to be printed in DOT format
		} else if (c == 'p') {
			pfsa = true;
		//determinize
		} else if (c == 'd') {
			det = true;
		//save as xml
		} else if (c == 'x') {
			xml = true;
		//minimize FSA
		} else if (c == 'm') {
			min = true;
		}
	}
	//if no RE or FSA is entered, print flags and exit
	if (!reg && !in) {
		//flags: MRwcmxdpefr:i:s:
		fprintf(stderr,"Usage: ./re flags\n");
		fprintf(stderr,"\t-i filename: read FSA from XML file\n");
		fprintf(stderr,"\t-r 'regexp': parse regular expression\n");
		fprintf(stderr,"\t-f         : create FSA from regular expression\n");
		fprintf(stderr,"\t-e         : remove epsilon-arcs from FSA\n");
		fprintf(stderr,"\t-d         : determinize FSA\n");
		fprintf(stderr,"\t-p         : print FSA (Graphviz format)\n");
		fprintf(stderr,"\t-w         : include weights with -p\n");
		fprintf(stderr,"\t-x         : save FSA as XML\n");
		fprintf(stderr,"\t-s string  : test string against FSA\n");
		fprintf(stderr,"\t-m         : minimize FSA\n");
		fprintf(stderr,"\t-M         : minimize FSA (Brzozowski)\n");
		fprintf(stderr,"\t-c         : take complement of FSA\n");
		fprintf(stderr,"\t-R         : reverse FSA\n");
		return 1;
	}
	//perform requisite operations depending on what flags are set
	//if RE has been entered
	if (reg) {
		//set up read-pointer and length variable
		theInputPtr = theInput;
		theInputLim = theInputPtr + strlen(theInput) + 1;
		//parse the regular expression
		yyparse();
		//release memory in that new variable.
		free(theInput);
		//print the regular expression parsed
		char *printable = printableRE(theRE);
		fprintf(stderr,"re: %s\n", printable);
		free(printable);
	//if an XML file has been specified
	} else if (in) {
		f = parseXML(filename);
	}
	//make automaton from RE if -f and -r flags are present
	if (fsa && theRE != NULL)
		f = makeFsa(theRE);
	//these flags only apply if there is an FSA
	if (f != NULL) {
		//remove epsilon arcs if -e
		if (eps)
			removeEpsilons(f);
		//determinize an undeterminized FSA if -d and -e
		if (eps && det && !f->determinized)
			determinize(f);
		//create complement of FSA
		if (comp && f->determinized)
			complement(f);
		//reverse FSA
		if (rev && f->determinized)
			reverse(f);
		//minimize a determinized FSA
		if (min && f->determinized)
			minimize(f);
		//minimize a determinized FSA by Brzozowski algorithm
		if (brzo && f->determinized && !min) {
			reverse(f);
			determinize(f);
			reverse(f);
			determinize(f);
		}
		//test a string if deterministic FSA exists and -s
		if (str && f->determinized) {
			if (testTheString(f, testString, weights))
				fprintf(stderr, "\"%s\" is legal.\n", testString);
			else
				fprintf(stderr, "\"%s\" is not legal.\n", testString);
		}
		//print the FSA in DOT format if -p and FSA exists
		if (pfsa)
			printFSA(f, weights);
		//print the FSA as XML if -x and FSA exists
		if (xml)
			printXML(f);
		//remove the FSA
		deallocFSA(f);
	}
	//remove the RE
	if (theRE != NULL)
		deallocRE(theRE);
	//remove the filename
	if (filename != NULL)
		free(filename);
	//remove the test string
	if (testString != NULL)
		free(testString);
	//for MallocDebug
/*
	while (1)
		;
	return 0;
*/
} 

